"""Configuration for the vocal burst model."""

from dataclasses import dataclass

from hume.models import ModelType
from hume.models.config.model_config_base import ModelConfigBase


@dataclass
class BurstConfig(ModelConfigBase["BurstConfig"]):
    """Configuration for the vocal burst model."""

    @classmethod
    def get_model_type(cls) -> ModelType:
        """Get the configuration model type.

        Returns:
            ModelType: Model type.
        """
        return ModelType.BURST



"""Base class for Hume clients."""

import logging
from importlib.metadata import version
from typing import ClassVar, Dict, Optional

from httpx import Client as HttpClient
from httpx import HTTPTransport, Response

from hume._common.protocol import Protocol
from hume._common.utilities.paging_utilities import Paging
from hume.error.hume_client_exception import HumeClientException

logger = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes
class ClientBase:
    """Base class for Hume API clients."""

    DEFAULT_PAGE_SIZE: ClassVar[int] = 50
    DEFAULT_RETRY_BACKOFF_FACTOR: ClassVar[int] = 2
    DEFAULT_RETRY_TIMEOUT: ClassVar[int] = 30 * 60
    DEFAULT_HTTP_RETRIES: ClassVar[int] = 3
    DEFAULT_HTTP_TIMEOUT: ClassVar[int] = 30
    DEFAULT_WS_OPEN_TIMEOUT: ClassVar[int] = 10
    DEFAULT_WS_CLOSE_TIMEOUT: ClassVar[int] = 10

    DEFAULT_API_VERSION: ClassVar[str] = "v0"
    DEFAULT_API_HTTP_BASE_URL: ClassVar[str] = "https://api.hume.ai"
    DEFAULT_API_WS_BASE_URI: ClassVar[str] = "wss://api.hume.ai"

    PAGING_LIMIT: ClassVar[int] = 100_000

    def __init__(
        self,
        api_key: str,
        *,
        page_size: int = DEFAULT_PAGE_SIZE,
        retry_backoff_factor: float = DEFAULT_RETRY_BACKOFF_FACTOR,
        retry_timeout: float = DEFAULT_RETRY_TIMEOUT,
        http_retries: int = DEFAULT_HTTP_RETRIES,
        http_timeout: float = DEFAULT_HTTP_TIMEOUT,
        open_timeout: int = DEFAULT_WS_OPEN_TIMEOUT,
        close_timeout: int = DEFAULT_WS_CLOSE_TIMEOUT,
        _api_version: str = DEFAULT_API_VERSION,
        _api_http_base_url: str = DEFAULT_API_HTTP_BASE_URL,
        _api_ws_base_uri: str = DEFAULT_API_WS_BASE_URI,
    ) -> None:
        """Construct a Hume API client.

        Args:
            api_key (str): Hume API key.

            page_size (int): Number of items to return per page.
            retry_backoff_factor (float): Backoff factor to use for exponential backoff during request retries.
            retry_timeout (float): Maximum time in seconds to retry a request.
            http_retries (int): Maximum number of times to retry a request.
            http_timeout (float): Time in seconds before canceling an HTTP request.
            open_timeout (int): Time in seconds before canceling a socket open operation.
            close_timeout (int): Time in seconds before canceling a socket close operation.
        """
        self._api_key = api_key
        self._page_size = page_size
        self._retry_backoff_factor = retry_backoff_factor
        self._retry_timeout = retry_timeout
        self._http_retries = http_retries
        self._http_timeout = http_timeout
        self._open_timeout = open_timeout
        self._close_timeout = close_timeout

        self._api_version = _api_version
        self._api_http_base_url = _api_http_base_url
        self._api_ws_base_uri = _api_ws_base_uri

        transport = HTTPTransport(http2=True, retries=self._http_retries)
        self._http_client = HttpClient(
            follow_redirects=True,
            timeout=http_timeout,
            transport=transport,
        )

    def _request(
        self,
        endpoint: str,
        *,
        method: str,
        body_json_str: Optional[str] = None,
        paging: Optional[Paging] = None,
    ) -> Response:
        headers = self._get_client_headers()

        if body_json_str is None:
            content = None
        else:
            headers["Content-Type"] = "application/json"
            content = body_json_str.encode(encoding="utf-8")

        params = {}
        if paging is not None:
            if paging.page_size is not None:
                params["page_size"] = paging.page_size
            if paging.page_number != 0:
                params["page_number"] = paging.page_number

        request = self._http_client.build_request(
            method,
            endpoint,
            content=content,
            params=params,
            headers=headers,
        )

        try:
            response = self._http_client.send(request)
            response.raise_for_status()
        except Exception as exc:  # pylint: disable=broad-exception-caught
            response_body = response.json()
            if "message" in response_body:
                raise HumeClientException(response_body["message"]) from exc
            raise HumeClientException(str(response_body)) from exc

        return response

    def _get_client_headers(self) -> Dict[str, str]:
        return {
            "X-Hume-Api-Key": self._api_key,
            "X-Hume-Client-Name": "python_sdk",
            "X-Hume-Client-Version": version("hume"),
        }

    def _build_endpoint(
        self,
        service: str,
        path: str,
        protocol: str = Protocol.HTTP,
    ) -> str:
        if protocol == Protocol.HTTP:
            base = self._api_http_base_url
        elif protocol == Protocol.WS:
            base = self._api_ws_base_uri

        return f"{base}/{self._api_version}/{service}/{path}"



"""Session settings model."""

from typing import Optional

from pydantic import BaseModel


class AudioSettings(BaseModel):
    """Audio settings model."""

    encoding: Optional[str] = "linear16"
    channels: Optional[int] = None
    sample_rate: Optional[int] = None


class SessionSettings(BaseModel):
    """Session settings model."""

    type: str = "session_settings"

    system_prompt: Optional[str] = None
    audio: Optional[AudioSettings] = None



from typing import Optional

from hume._common.utilities.model_utilities import BaseModel


class Paging(BaseModel):
    """HTTP response paging parameters."""

    page_size: Optional[int]
    page_number: int = 0



"""API request and response models for EVI configurations."""

from typing import List, Optional

from hume._common.utilities.model_utilities import BaseModel
from hume._voice.models.tools_models import ToolMeta


class PromptResponse(BaseModel):
    """Response model for an EVI prompt."""

    id: str
    version: int
    version_description: Optional[str]
    name: str
    created_on: int
    modified_on: int
    text: Optional[str]


class PromptsResponse(BaseModel):
    """Response model for a page of EVI prompts."""

    page_number: int
    page_size: int
    prompts: List[PromptResponse]


class PromptMeta(BaseModel):
    """Prompt metadata."""

    id: str
    version: int


class LanguageModelConfig(BaseModel):
    """Language model configuration for EVI."""

    model_provider: str
    model_resource: str
    temperature: Optional[float] = None


class PostPromptRequest(BaseModel):
    """Post request model for creating a new EVI prompt."""

    name: str
    version_description: Optional[str]
    text: Optional[str]


class ConfigResponse(BaseModel):
    """Response model for an EVI configurations."""

    id: str
    version: int
    version_description: Optional[str]
    name: str
    created_on: int
    modified_on: int
    prompt: Optional[PromptResponse]


class ConfigsResponse(BaseModel):
    """Response model for a page of EVI configurations."""

    page_number: int
    page_size: int
    configs_page: List[ConfigResponse]


class VoiceIdentityConfig(BaseModel):
    """Configuration for changing the voice of EVI."""

    provider: Optional[str] = None
    name: Optional[str] = None


class BuiltinToolConfig(BaseModel):
    """Configuration for a built-in EVI tool."""

    name: str
    tool_type: str
    fallback_content: Optional[str]


class PostConfigRequest(BaseModel):
    """Post request model for creating a new EVI configuration."""

    name: str
    version_description: Optional[str]
    prompt: PromptMeta
    voice: Optional[VoiceIdentityConfig]
    language_model: Optional[LanguageModelConfig]
    tools: Optional[List[ToolMeta]]


class ConfigMeta(BaseModel):
    """EVI configuration metadata."""

    id: Optional[str]
    version: Optional[int]


class VoiceConfig(BaseModel):
    """EVI configuration."""

    id: str
    name: str
    description: Optional[str]
    created_on: int
    modified_on: int
    # TODO: Add tool info
    prompt: Optional[str]



"""Batch job state."""

from dataclasses import dataclass
from typing import Optional

from hume._measurement.batch.batch_job_status import BatchJobStatus


@dataclass
class BatchJobState:
    """Batch job state.

    Args:
        status (BatchJobStatus): Status of the batch job.
        created_timestamp_ms (Optional[int]): Time when job was created.
        started_timestamp_ms (Optional[int]): Time when job started.
        ended_timestamp_ms (Optional[int]): Time when job ended.
    """

    status: BatchJobStatus
    created_timestamp_ms: Optional[int]
    started_timestamp_ms: Optional[int]
    ended_timestamp_ms: Optional[int]



"""Sender for streaming audio from a microphone."""

import json
import logging
from dataclasses import dataclass
from typing import Protocol

from hume._voice.microphone.microphone import Microphone
from hume._voice.voice_socket import VoiceSocket

logger = logging.getLogger(__name__)


class Sender(Protocol):
    """Protocol for sending streaming audio to an EVI connection."""

    async def on_audio_begin(self) -> None:
        """Handle the start of an audio stream."""
        raise NotImplementedError()

    async def on_audio_end(self) -> None:
        """Handle the end of an audio stream."""
        raise NotImplementedError()

    async def send(self, *, socket: VoiceSocket) -> None:
        """Send audio data over an EVI socket.

        Args:
            socket (VoiceSocket): EVI socket.
        """
        raise NotImplementedError()

    async def send_tool_response(self, *, socket: VoiceSocket, tool_call_id: str, content: str) -> None:
        """Send a tool response over an EVI socket.

        Args:
            socket (VoiceSocket): EVI socket.
            tool_call_id (str): Tool call ID.
            content (str): Tool response content.
        """
        raise NotImplementedError()


@dataclass
class MicrophoneSender(Sender):
    """Sender for streaming audio from a microphone."""

    microphone: Microphone
    send_audio: bool
    allow_interrupt: bool

    @classmethod
    def new(cls, *, microphone: Microphone, allow_interrupt: bool) -> "MicrophoneSender":
        """Create a new microphone sender.

        Args:
            microphone (Microphone): Microphone instance.
            allow_interrupt (bool): Whether to allow interrupting the audio stream.
        """
        return cls(microphone=microphone, send_audio=True, allow_interrupt=allow_interrupt)

    async def on_audio_begin(self) -> None:
        """Handle the start of an audio stream."""
        self.send_audio = self.allow_interrupt

    async def on_audio_end(self) -> None:
        """Handle the end of an audio stream."""
        self.send_audio = True

    async def send(self, *, socket: VoiceSocket) -> None:
        """Send audio data over an EVI socket.

        Args:
            socket (VoiceSocket): EVI socket.
        """
        async for byte_str in self.microphone:
            if self.send_audio:
                await socket.send(byte_str)

    async def send_tool_response(self, *, socket: VoiceSocket, tool_call_id: str, content: str) -> None:
        """Send a tool response over an EVI socket.

        Args:
            socket (VoiceSocket): EVI socket.
            tool_call_id (str): Tool call ID.
            content (str): Tool response content.
        """
        response_message = {
            "type": "tool_response",
            "tool_call_id": tool_call_id,
            "content": content,
        }
        await socket.send(json.dumps(response_message).encode("utf-8"))



"""Streaming API client."""

from contextlib import asynccontextmanager
from typing import Any, AsyncIterator, Optional, Sequence

import websockets

from hume._common.client_base import ClientBase
from hume._common.protocol import Protocol
from hume._common.utilities.config_utilities import deserialize_configs
from hume._measurement.stream.stream_socket import StreamSocket
from hume.error.hume_client_exception import HumeClientException
from hume.models.config.model_config_base import ModelConfigBase


class HumeStreamClient(ClientBase):
    """Streaming API client.

    Example:
        ```python
        import asyncio

        from hume import HumeStreamClient
        from hume.models.config import BurstConfig
        from hume.models.config import ProsodyConfig

        async def main():
            client = HumeStreamClient("<your-api-key>")
            configs = [BurstConfig(), ProsodyConfig()]
            async with client.connect(configs) as socket:
                result = await socket.send_file("<your-audio-filepath>")
                print(result)

        asyncio.run(main())
        ```
    """

    @asynccontextmanager
    async def connect(
        self,
        configs: Sequence[ModelConfigBase],
        stream_window_ms: Optional[int] = None,
    ) -> AsyncIterator[StreamSocket]:
        """Connect to the streaming API.

        Note: Only one config per model type should be passed.
            If more than one config is passed for a given model type, only the last config will be used.

        Args:
            configs (List[ModelConfigBase]): List of job configs.
            stream_window_ms (Optional[int]): Length of the sliding window in milliseconds to use when
                aggregating media across streaming payloads within one WebSocket connection.
        """
        endpoint = self._build_endpoint("stream", "models", protocol=Protocol.WS)
        try:
            # pylint: disable=no-member
            async with websockets.connect(  # type: ignore[attr-defined]
                endpoint,
                extra_headers=self._get_client_headers(),
                close_timeout=self._close_timeout,
                open_timeout=self._open_timeout,
            ) as protocol:
                yield StreamSocket(protocol, list(configs), stream_window_ms=stream_window_ms)
        except websockets.exceptions.InvalidStatusCode as exc:
            status_code: int = exc.status_code
            if status_code == 401:  # Unauthorized
                message = "HumeStreamClient initialized with invalid API key."
                raise HumeClientException(message) from exc
            raise HumeClientException("Unexpected error when creating streaming connection") from exc

    @asynccontextmanager
    async def _connect_with_configs_dict(self, configs_dict: Any) -> AsyncIterator[StreamSocket]:
        """Connect to the streaming API with a single models configuration dict.

        Args:
            configs_dict (Any): Models configurations dict. This should be a dict from model name
                to model configuration dict. An empty dict uses the default configuration.
        """
        configs = deserialize_configs(configs_dict)
        async with self.connect(configs) as websocket:
            yield websocket
